"""
Plotting backends for Owl.

Copyright &copy; 2025 - Martin-D. Lacasse

Disclaimers: This code is for educatonal purposes only and does not constitute financial advice.

"""

from .factory import PlotFactory

__all__ = ['PlotFactory']
